#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
	The protection device (ASIC27) has access to the 68K ROM (main program ROM)
	and hash checks it before it before writing the decryption table to a RAM
	region (see kovsh_tab) that is used to decrypt the main program ROM.
	If the hash fails, the ASIC27 will write junk values to the decryption RAM.
*/

struct define_hash
{
	unsigned int start_address; // where does this hash region start?
	unsigned short add_or_xor; // add or xor bytes depending on a configuration table in the ASIC27 ROM
	unsigned short length; // how long is the region we're hashing
};

static define_hash hash_regions_kovsh[0x0d] = {	// table in asic27 rom @ $2144
	{ 0x00000, 0x0002, 0x1000 },
	{ 0x48000, 0x0001, 0x8000 },
	{ 0x58000, 0x0002, 0x8000 },
	{ 0x68000, 0x0001, 0x8000 },
	{ 0x78000, 0x0002, 0x8000 },
	{ 0x88000, 0x0001, 0x8000 },
	{ 0x98000, 0x0002, 0x8000 },
	{ 0xa8000, 0x0001, 0x8000 },
	{ 0xb8000, 0x0002, 0x8000 },
	{ 0xc8000, 0x0001, 0x8000 },
	{ 0xd8000, 0x0002, 0x4000 },
	{ 0xe8000, 0x0001, 0x3000 },
	{ 0x39abe, 0, 0 }	// address to find stored hash
};

static void verify_hash_function(unsigned char *src)
{
	int i, j;
	unsigned short shash = 0, hash = 0, value, *rom = (unsigned short*)src;
	define_hash *ptr = hash_regions_kovsh;
	int entries = sizeof(hash_regions_kovsh)/sizeof(define_hash);

	for (i = 0; i < entries-1; i++, hash += shash)
	{
		for (j = 0, shash = 0; j < ptr->length; j+=2)
		{
			value = rom[(ptr->start_address + j)/2];
			shash = (ptr->add_or_xor == 1) ? (shash + value) : (shash ^ value);
		}
		ptr++;
	}
	
	printf ("Calculated: %4.4x, Correct: %4.4x\n", hash, *((unsigned short*)(src + ptr->start_address)));

	// apply calculated hash to rom (after encryption)
	*((unsigned short*)(src + ptr->start_address)) = hash;
}

static const unsigned char kovsh_tab[256] = {
	0xe7, 0x06, 0xa3, 0x70, 0xf2, 0x58, 0xe6, 0x59, 0xe4, 0xcf, 0xc2, 0x79, 0x1d, 0xe3, 0x71, 0x0e,
	0xb6, 0x90, 0x9a, 0x2a, 0x8c, 0x41, 0xf7, 0x82, 0x9b, 0xef, 0x99, 0x0c, 0xfa, 0x2f, 0xf1, 0xfe,
	0x8f, 0x70, 0xf4, 0xc1, 0xb5, 0x3d, 0x7c, 0x60, 0x4c, 0x09, 0xf4, 0x2e, 0x7c, 0x87, 0x63, 0x5f,
	0xce, 0x99, 0x84, 0x95, 0x06, 0x9a, 0x20, 0x23, 0x5a, 0xb9, 0x52, 0x95, 0x48, 0x2c, 0x84, 0x60,
	0x69, 0xe3, 0x93, 0x49, 0xb9, 0xd6, 0xbb, 0xd6, 0x9e, 0xdc, 0x96, 0x12, 0xfa, 0x60, 0xda, 0x5f,
	0x55, 0x5d, 0x5b, 0x20, 0x07, 0x1e, 0x97, 0x42, 0x77, 0xea, 0x1d, 0xe0, 0x70, 0xfb, 0x6a, 0x00,
	0x77, 0x9a, 0xef, 0x1b, 0xe0, 0xf9, 0x0d, 0xc1, 0x2e, 0x2f, 0xef, 0x25, 0x29, 0xe5, 0xd8, 0x2c,
	0xaf, 0x01, 0xd9, 0x6c, 0x31, 0xce, 0x5c, 0xea, 0xab, 0x1c, 0x92, 0x16, 0x61, 0xbc, 0xe4, 0x7c,
	0x5a, 0x76, 0xe9, 0x92, 0x39, 0x5b, 0x97, 0x60, 0xea, 0x57, 0x83, 0x9c, 0x92, 0x29, 0xa7, 0x12,
	0xa9, 0x71, 0x7a, 0xf9, 0x07, 0x68, 0xa7, 0x45, 0x88, 0x10, 0x81, 0x12, 0x2c, 0x67, 0x4d, 0x55,
	0x33, 0xf0, 0xfa, 0xd7, 0x1d, 0x4d, 0x0e, 0x63, 0x03, 0x34, 0x65, 0xe2, 0x76, 0x0f, 0x98, 0xa9,
	0x5f, 0x9a, 0xd3, 0xca, 0xdd, 0xc1, 0x5b, 0x3d, 0x4d, 0xf8, 0x40, 0x08, 0xdc, 0x05, 0x38, 0x00,
	0xcb, 0x24, 0x02, 0xff, 0x39, 0xe2, 0x9e, 0x04, 0x9a, 0x08, 0x63, 0xc8, 0x2b, 0x5a, 0x34, 0x06,
	0x62, 0xc1, 0xbb, 0x8a, 0xd0, 0x54, 0x4c, 0x43, 0x21, 0x4e, 0x4c, 0x99, 0x80, 0xc2, 0x3d, 0xce,
	0x2a, 0x7b, 0x09, 0x62, 0x1a, 0x91, 0x9b, 0xc3, 0x41, 0x24, 0xa0, 0xfd, 0xb5, 0x67, 0x93, 0x07,
	0xa7, 0xb8, 0x85, 0x8a, 0xa1, 0x1e, 0x4f, 0xb6, 0x75, 0x38, 0x65, 0x8a, 0xf9, 0x7c, 0x00, 0xa0
};

void pgm_decrypt_kovsh(unsigned char *rom)
{
	int i;
	unsigned short *src = (unsigned short *)rom;

	for (i = 0; i < 0x200000/2; i++) {
		unsigned short x = src[i];

		if ((i & 0x040080) != 0x000080) x ^= 0x0001;
		if ((i & 0x004008) == 0x004008 && (i & 0x180000) != 0x000000) x ^= 0x0002;
		if ((i & 0x000030) == 0x000010) x ^= 0x0004;
		if ((i & 0x000242) != 0x000042) x ^= 0x0008;
		if ((i & 0x008100) == 0x008000) x ^= 0x0010;
		if ((i & 0x002004) != 0x000004) x ^= 0x0020;
		if ((i & 0x011800) != 0x010000) x ^= 0x0040;
		if ((i & 0x000820) == 0x000820) x ^= 0x0080;

		x ^= kovsh_tab[i & 0xff] << 8;

		src[i] = x;
	}
}

int main()
{
	unsigned char *rom = (unsigned char *)malloc(0x400000);
	FILE *fa = fopen("ddp_j.bin", "rb");
	if (fa == NULL) {
		free (rom);
		printf ("ddp_j.bin not found!\n");
		return 1;
	}
	fread (rom, 0x400000, 1, fa);
	fclose (fa);

	memset (rom + 0x200000, 0xff, 0x200000);

	pgm_decrypt_kovsh(rom);
	verify_hash_function(rom);

	fa = fopen("ddp_j_encrypted.bin", "wb");
	fwrite (rom, 0x400000, 1, fa);
	fclose (fa);

	free (rom);
}
